@@ -1,4 +1,4 @@
-/* 02dec20 Software Lab. Alexander Burger */
+/* 21nov21 Software Lab. Alexander Burger */
 
 var FormReq = new XMLHttpRequest();
 FormReq.upload.addEventListener("progress", dropProgress, false);
@@ -260,19 +260,7 @@ function post(form, auto, exe, file) {
                   switch (j) {
 
                   case 0:  // '#'
-                     var cls;
-
-                     val = txt[i++].substr(1);
-                     if ((cls = fld.getAttribute("class")) != null) {
-                        j = cls.indexOf("  ");
-                        if (!val)
-                           val = j >= 0? cls.substr(j+2) : "";
-                        else if (j >= 0)
-                           val += cls.substr(j);
-                        else
-                           val += "  " + cls;
-                     }
-                     fld.setAttribute("class", val);
+                     fld.setAttribute("class", txt[i++].substr(1));
                      break;
 
                   case 1:  // '*'
@@ -312,7 +300,7 @@ function post(form, auto, exe, file) {
          var a = Queue.shift();
          post(a[0], a[1], a[2], a[3]);
       }
-   }
+   };
    if (!exe)
       data = "";
    else {
